/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/01/14
 */
package net.cattaka.converttosculpt.data;


/**
 * @author cattaka
 * ベクトルの各種演算を行うためのクラス。
 */
public class BMath
{
    /**
     * 球状マップの計算を行う
     * 
     * @param dest          出力先
     * @param longitudes    経度方向の分割数
     * @param latitudes     緯度方向の分割数
     * @param meshData      メッシュデータ
     * @param pos           中心座標
     */
    public static void calcSphereMap(BVector3d[][] dest, int longitudes, int latitudes, MeshData meshData, BVector3d pos, BVector3d axisAngle) {

        BQuaternion4d quat = new BQuaternion4d();
        {
            BQuaternion4d qx = new BQuaternion4d();
            BQuaternion4d qy = new BQuaternion4d();
            BQuaternion4d qz = new BQuaternion4d();
            quat.setAxisAngle(new BVector3d(1,0,0), axisAngle.x);
            qy.setAxisAngle(new BVector3d(0,1,0), axisAngle.y);
            qz.setAxisAngle(new BVector3d(0,0,1), axisAngle.z);
            quat.mul(qy);
            quat.mul(qz);
        }
        
        BVector3d dir = new BVector3d();
        BVector3d tMin = new BVector3d();
        BVector3d tMax = new BVector3d();
        
        double rangeLon;
        double rangelati;
        rangeLon = (double)longitudes;
        
        rangelati = (double)(latitudes-1);
        
        for (int i=0;i < longitudes;i++) {
            double lon = (double)i/ rangeLon;
            lon = lon * 2 * Math.PI;
            for (int j=0;j < latitudes;j++) {
                double lati = (double)j / rangelati;
                lati = lati - 0.5;
                lati = lati * Math.PI;
                
                dir.x = Math.cos(lon) * Math.cos(lati);
                dir.y = Math.sin(lon) * Math.cos(lati);
                dir.z = Math.sin(lati);
                
                quat.rotate(dir);
                
                calcIntersectMinMaxPoint(tMin, tMax, meshData, pos, dir);
                dest[i][j].set(tMax);
            }
        }
    }
    
    
    /**
     * メッシュデータと直線の交点を求め、原点より最も遠い2点を得る。
     * ただし、交差しない場合は原点を取る。
     * 
     * @param destMin   交点の出力先(最小)
     * @param destMax   交点の出力先(最大)
     * @param meshData  メッシュデータ
     * @param pos       直線の原点
     * @param dir       直線の方向
     */
    public static void calcIntersectMinMaxPoint(BVector3d destMin, BVector3d destMax, MeshData meshData, BVector3d pos, BVector3d dir) {
        
        BVector3d[] css = meshData.getCoordinate();
        short[][] cis = meshData.getCoordIndex();
        
        BVector3d[] pg = new BVector3d[4];
        BVector3d tmp = new BVector3d();
        double minDist = 0;
        double maxDist = 0;
        
        boolean detectedFlag = false;
        for (int i=0;i<cis.length;i++) {
            int j;
            for (j=0;j<cis[i].length;j++) {
                if (j<pg.length) {
                    pg[j] = css[cis[i][j]];
                } else {
                    throw new RuntimeException("Invalid polygon.");
                }
            }
            if (calcIntersectPoint(tmp, pg, j, pos, dir)) {
                double distance = BVector3d.getDot(dir,tmp);
                if (detectedFlag) {
                    if (distance > maxDist) {
                        maxDist = distance;
                        destMax.set(tmp);
                    }
                    if (distance < minDist) {
                        minDist = distance;
                        destMin.set(tmp);
                    }
                } else {
                    maxDist = distance;
                    destMax.set(tmp);
                    minDist = distance;
                    destMin.set(tmp);
                    detectedFlag = true;
                }
            }
        }
        if (!detectedFlag) {
            destMin.set(pos);
            destMax.set(pos);
        }
    }

    /**
     * 多角形と直線の交点を求める。
     * 現時点では頂点数が３，４のみサポート。
     * 
     * @param dest  交点の出力先
     * @param pg    多角形の頂点の配列
     * @param pos   直線の原点
     * @param dir   直線の方向
     * @return      交差している場合はtrue、そうでなければfalseを返す。
     */
    public static boolean calcIntersectPoint(BVector3d dest, BVector3d[] pg, int num, BVector3d pos, BVector3d dir) {
        if (num == 3 && pg.length >= 3) {
            return calcIntersectPoint(dest, pg[0], pg[1], pg[2], pos, dir);
        } else if (num == 4 && pg.length >= 4){
            if (calcIntersectPoint(dest, pg[0], pg[1], pg[2], pos, dir)) {
                return true;
            } else if (calcIntersectPoint(dest, pg[2], pg[3], pg[0], pos, dir)) {
                return true;
            } else {
                return false;
            }
        } else {
            throw new RuntimeException("Invalid polygon : " + num);
        }
    }

    
    /**
     * 三角形と直線の交点を求める。
     * 
     * @param dest  交点の出力先
     * @param t1    三角形の頂点1
     * @param t2    三角形の頂点2
     * @param t3    三角形の頂点3
     * @param pos   直線の原点
     * @param dir   直線の方向
     * @return      交差している場合はtrue、そうでなければfalseを返す。
     */
    public static boolean calcIntersectPoint(BVector3d dest, BVector3d t1, BVector3d t2, BVector3d t3, BVector3d pos, BVector3d dir) {
        BVector3d tmp1 = new BVector3d();
        BVector3d tmp2 = new BVector3d();
        BVector3d tmp3 = new BVector3d();

        //三角形の内部かどうか判定
        double d1,d2,d3;
        tmp1.set(t1);
        tmp1.sub(pos);
        tmp2.set(t2);
        tmp2.sub(pos);
        BVector3d.getCross(tmp3, tmp1, tmp2);
        d1 = BVector3d.getDot(tmp3, dir);
        tmp1.set(t2);
        tmp1.sub(pos);
        tmp2.set(t3);
        tmp2.sub(pos);
        BVector3d.getCross(tmp3, tmp1, tmp2);
        d2 = BVector3d.getDot(tmp3, dir);
        tmp1.set(t3);
        tmp1.sub(pos);
        tmp2.set(t1);
        tmp2.sub(pos);
        BVector3d.getCross(tmp3, tmp1, tmp2);
        d3 = BVector3d.getDot(tmp3, dir);
        if (!((d1>=0 && d2>=0 && d3>=0) || (d1<=0 && d2<=0 && d3<=0))) {
            //三角形と交差していない場合
            return false;
        }
        if (d1==0 && d2==0 && d3==0) {
            //3点が1点に重なっている場合
            return false;
        }
        
        
        //面の４次元ベクトルを作成
        //3Dゲームプログラマーのための数学：6.3章参照
        BVector3d v12 = new BVector3d(t2);
        BVector3d v23 = new BVector3d(t3);
        v12.sub(t1);
        v23.sub(t2);
        BVector4d plane = new BVector4d();
        BVector3d n = new BVector3d();

        BVector3d.getCross(n, v12, v23);
        if (n.isZero()) {
            //法線が零ベクトルの場合、無理
            return false;
        }
        
        plane.a = n.x;
        plane.b = n.y;
        plane.c = n.z;
        plane.d = - BVector3d.getDot(n, t1);
        
        return calcIntersectPoint(dest, plane, pos, dir);
    }
    
    
    /**
     * 面と直線の交点を求める。
     * 
     * @param dest  交点の出力先
     * @param plane 面
     * @param pos   直線の原点
     * @param dir   直線の方向
     * @return      交差している場合はtrue、そうでなければfalseを返す。
     */
    public static boolean calcIntersectPoint(BVector3d dest, BVector4d plane, BVector3d pos, BVector3d dir) {
        //3Dゲームプログラマーのための数学：6.2章参照
        double t1, t2;
        t1 = plane.a * pos.x + plane.b * pos.y + plane.c * pos.z + plane.d;
        if (t1 == 0) {
            //原点が面上に存在している
            dest.set(pos);
            return true;
        }
        
        t2 = plane.a * dir.x + plane.b * dir.y + plane.c * dir.z;
        if (t2 == 0) {
            //方向が面に平行である
            return false;
        }
        
        double t = - t1 / t2;
        dest.set(pos);
        dest.scaleAdd(t, dir);
        return true;
    }
}
