/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2008/02/16
 */
package net.cattaka.converttosculpt.data;

import net.cattaka.converttosculpt.exception.InvalidMeshDataException;

/**
 * メッシュデータを保持するクラス。
 * 
 * @author cattaka
 */

public class MeshData
{
    private String name;
    private short[][] coordIndex;
    private short[][] texCoordIndex;
    private BVector3d[] coordinate;
    private BVector3d[] texCoordinate;
    
    public MeshData()
    {
    }

    /**
     * Copy constructor.
     * @param src
     */
    public MeshData(MeshData src) {
        if (src == null) {
            throw new NullPointerException();
        }
        this.set(src);
    }
    
    /**
     * Create new MeshData.
     * Warning this constructor does not check validity.
     * 
     * @param name
     * @param coordIndex
     * @param coordinate
     * @param texCoordIndex
     * @param texCoordinate
     */
    public MeshData(String name, short[][] coordIndex, BVector3d[] coordinate, short[][] texCoordIndex, BVector3d[] texCoordinate) {
        this.name = name;
        this.coordIndex = coordIndex.clone();
        this.texCoordIndex = texCoordIndex.clone();
        
        BVector3d[] cs = coordinate;
        this.coordinate = new BVector3d[cs.length];
        for (int i=0;i<cs.length;i++) {
            this.coordinate[i] = new BVector3d(cs[i]);
        }

        BVector3d[] tcs = texCoordinate;
        this.texCoordinate = new BVector3d[tcs.length];
        for (int i=0;i<tcs.length;i++) {
            this.texCoordinate[i] = new BVector3d(tcs[i]);
        }
    }
    
    /**
     * 
     * 
     * @param src
     */
    public void set(MeshData src) {
        this.name = src.getName();
        if (src.getCoordIndex() != null) {
            this.coordIndex = src.getCoordIndex().clone();
        }
        if (src.getTexCoordIndex() != null) {
            this.texCoordIndex = src.getTexCoordIndex().clone();
        }
        
        BVector3d[] cs = src.getCoordinate();
        if (cs != null) {
            this.coordinate = new BVector3d[cs.length];
            for (int i=0;i<cs.length;i++) {
                if (cs[i] != null) {
                    this.coordinate[i] = new BVector3d(cs[i]);
                } else {
                    this.coordinate[i] = null;
                }
            }
        }

        BVector3d[] tcs = src.getTexCoordinate();
        if (tcs != null) {
            this.texCoordinate = new BVector3d[tcs.length];
            for (int i=0;i<tcs.length;i++) {
                if (tcs[i] != null) {
                    this.texCoordinate[i] = new BVector3d(tcs[i]);
                } else {
                    this.texCoordinate[i] = null;
                }
            }
        }
    }
    
    public String getName()
    {
        return name;
    }

    public void setName(String name)
    {
        this.name = name;
    }

    public short[][] getCoordIndex()
    {
        return coordIndex;
    }

    public short[][] getTexCoordIndex()
    {
        return texCoordIndex;
    }

    public BVector3d[] getCoordinate()
    {
        return coordinate;
    }

    public BVector3d[] getTexCoordinate()
    {
        return texCoordinate;
    }

    public void setCoordIndex(short[][] coordIndex)
    {
        this.coordIndex = coordIndex;
    }

    public void setTexCoordIndex(short[][] texCoordIndex)
    {
        this.texCoordIndex = texCoordIndex;
    }

    public void setCoordinate(BVector3d[] coordinate)
    {
        this.coordinate = coordinate;
    }

    public void setTexCoordinate(BVector3d[] texCoordinate)
    {
        this.texCoordinate = texCoordinate;
    }
    /**
     * メッシュデータの内容をチェックする。
     * この関数が例外を発生させなかった場合、次のことを保証する。
     *  coordinateが存在する。
     *  coordinateにnullが含まれない。
     *  coordIndexが存在する。
     *  coordIndexにnullが含まれない。
     *  coordIndexの範囲が正しい。
     *  texCoordIndexが存在する場合:
     *      texCoordIndexとcoordIndexの長さが等しい。
     *      texCoordinateが存在する。
     *      texCoordIndexの範囲が正しい。
     *      ※:texCoordIndexにはnullが含まれる事がある。
     * @throws InvalidMeshDataException
     */
    public void isValid() throws InvalidMeshDataException {
        if (coordinate == null) {
            throw new InvalidMeshDataException("coordinate is null");
        }
        for (int i=0;i<this.coordinate.length;i++) {
            if (coordinate[i] == null) {
                throw new InvalidMeshDataException("coordinate containes null : index = " + i);
            }
        }
        if (coordIndex == null) {
            throw new InvalidMeshDataException("coordIndex is null");
        }
        for (int i=0;i<coordIndex.length;i++) {
            short[] ci = coordIndex[i];
            if (ci == null) {
                throw new InvalidMeshDataException("coordIndex containes null : index = " + i);
            }
            for (int j=0;j<ci.length;j++) {
                if (ci[j] < 0 || coordinate.length <= ci[j]) {
                    throw new InvalidMeshDataException("coordIndex("+i+"," + j +") is out of bounds of coordinate : " + ci[j]);
                }
            }
        }
        
        if (texCoordIndex != null) {
            if (texCoordIndex.length != coordIndex.length) {
                throw new InvalidMeshDataException("length of coordIndex and texCoordIndex is mismatch");
            }
            if (texCoordinate == null) {
                throw new InvalidMeshDataException("texCoordinate is null");
            }
            for (int i=0;i<texCoordIndex.length;i++) {
                short[] tci = texCoordIndex[i];
                if (tci == null) {
                    continue;
                }
                for (int j=0;j<tci.length;j++) {
                    if (tci[j] < 0 || texCoordinate.length <= tci[j]) {
                        throw new InvalidMeshDataException("texCoordIndex("+i+"," + j +") is out of bounds of coordinate : " + tci[j]);
                    }
                }
            }
        }

    }
}
