/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2008/03/22
 */
package net.cattaka.converttosculpt.logic;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import net.cattaka.converttosculpt.data.BVector3d;
import net.cattaka.converttosculpt.data.MeshData;
import net.cattaka.converttosculpt.exception.InvalidMeshDataException;
import net.cattaka.converttosculpt.exception.InvalidObjException;


public class MeshConverterObj implements MeshConverter
{
    
    public void writeMeshData(MeshData meshData, OutputStream out) throws IOException,InvalidMeshDataException {
        String objStr = convertMeshDataToObj(meshData);
        
        OutputStreamWriter writer = new OutputStreamWriter(out);
        writer.write(objStr);
        writer.flush();
   }

    
    /**
     * メッシュデータをObj形式の文字列に変換する
     * 
     * @return
     */
    public static String convertMeshDataToObj(MeshData meshData) {
        //テキストデータ作成
        String name = meshData.getName();
        BVector3d[] coordinate = meshData.getCoordinate();
        BVector3d[] texCoordinate = meshData.getTexCoordinate();
        short[][] coordIndex = meshData.getCoordIndex();
        short[][] texCoordIndex = meshData.getTexCoordIndex();
        
        DecimalFormat df = new DecimalFormat();
        df.setMinimumFractionDigits(6);
        df.setMaximumFractionDigits(6);
        
        StringBuilder sb = new StringBuilder(1024*1024);
        sb.append("# ConvertToSculpt build " + VersionBundle.getBuildNumber() + "\n");
        sb.append("# www.cattaka.net\n");
        sb.append("o " + name + "\n");

        //coordinate作成
        for (int i=0;i<coordinate.length;i++) {
            sb.append("v ");
            BVector3d v = coordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
            sb.append(" ");
            sb.append(df.format(v.z));
            sb.append("\n");
        }
    
        //texCoordinate作成
        for (int i=0;i<texCoordinate.length;i++) {
            sb.append("vt ");
            BVector3d v = texCoordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
            sb.append(" ");
            sb.append(df.format(v.z));
            sb.append("\n");
        }
        
        //coordIndex/texCoordIndex作成
        for (int i=0;i<coordIndex.length;i++) {
            sb.append("f ");
            short[] ci = coordIndex[i];
            short[] tci = texCoordIndex[i];
            for (int j=0;j<ci.length;j++) {
                if (j > 0) {
                    sb.append(" ");
                }
                sb.append((ci[j] + 1));
                sb.append("/");
                sb.append((tci[j] + 1));
            }
            sb.append("\n");
        }
        
        return sb.toString();
    }

    /**
     * 
     * @param in       入力
     * @param result    出力
     * @param messages  出力
     * @throws InvalidObjException
     * @throws IOException
     */
    public void readMeshData(InputStream in, List<MeshData> result, List<String> messages) throws IOException,InvalidMeshDataException {
        BufferedReader reader = new BufferedReader(new InputStreamReader(in));
        ArrayList<short[]> coordIndexList = new ArrayList<short[]>(2048);
        ArrayList<short[]> texCoordIndexList = new ArrayList<short[]>(2048);
        ArrayList<BVector3d> coordinateList = new ArrayList<BVector3d>(2048);
        ArrayList<BVector3d> texCoordinateList = new ArrayList<BVector3d>(2048);

        //Obj形式はオブジェクトごとではなく、全体を通してユニークな頂点番号をcoordIndexに割り当てるため、
        //そのオフセットの計算用
        int vCount = 0;
        int vtCount = 0;
        int vOffset = 0;
        int vtOffset = 0;
        
        boolean firstTimeFlag = true;
        String src;
        String name = null;
        int lineNum = 0;
        do {
            src = reader.readLine();
            lineNum++;
            String[] tokens = null;
            if (src != null) {
                tokens = src.split("\\s+", 2);
            }
            try {
                if (tokens == null || tokens[0].equals("o")) {
                    if (!firstTimeFlag) {
                        MeshData meshData = new MeshData();
                        meshData.setName(name);
                        meshData.setCoordinate(coordinateList.toArray(new BVector3d[coordinateList.size()]));
                        meshData.setCoordIndex(coordIndexList.toArray(new short[coordIndexList.size()][]));
                        if (texCoordinateList.size() > 0
                                && texCoordIndexList.size() > 0) {
                            meshData.setTexCoordinate(texCoordinateList.toArray(new BVector3d[texCoordinateList.size()]));
                            meshData.setTexCoordIndex(texCoordIndexList.toArray(new short[texCoordIndexList.size()][]));
                        }
                        result.add(meshData);
                    } else {
                        firstTimeFlag = false;
                    }
                    if (tokens == null) {
                        break;
                    }
                    if (tokens.length < 2) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }

                    name = tokens[1];
                    coordinateList.clear();
                    texCoordinateList.clear();
                    coordIndexList.clear();
                    texCoordIndexList.clear();
                    vOffset = vCount;
                    vtOffset = vtCount;
                } else if (tokens[0].equals("v")) {
                    if (tokens.length != 2) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }
                    String[] tokens2 = tokens[1].split("\\s+");
                    if (tokens2.length < 3) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }
                    double x = Double.parseDouble(tokens2[0]);
                    double y = Double.parseDouble(tokens2[1]);
                    double z = Double.parseDouble(tokens2[2]);
                    coordinateList.add(new BVector3d(x,y,z));
                    vCount++;
                } else if (tokens[0].equals("vt")) {
                    if (tokens.length != 2) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }
                    String[] tokens2 = tokens[1].split("\\s+");
                    if (tokens2.length < 2) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }
                    double x = Double.parseDouble(tokens2[0]);
                    double y = Double.parseDouble(tokens2[1]);
                    texCoordinateList.add(new BVector3d(x,y,0));
                    vtCount++;
                } else if (tokens[0].equals("f")) {
                    if (tokens.length != 2) {
                        throw new InvalidObjException("invalid line (" + lineNum + "): " + src);
                    }
                    String[] tokens2 = tokens[1].split("\\s+");
                    short[] ci = new short[tokens2.length];
                    short[] tci = new short[tokens2.length];
                    boolean enableTex = true;
                    for (int i=0;i<tokens2.length;i++) {
                        String[] strs = tokens2[i].split("/");
                        ci[i]  = (short)(Short.parseShort(strs[0]) - 1 - vOffset);
                        if (strs.length > 1) {
                            try {
                                tci[i] = (short)(Short.parseShort(strs[1]) - 1 - vtOffset);
                            } catch (NumberFormatException e) {
                                //１つでもtexCoordIndexが無かったらテクスチャが無いものとする。
                                enableTex = false;
                            }
                        } else {
                            //１つでもtexCoordIndexが無かったらテクスチャが無いものとする。
                            enableTex = false;
                        }
                    }
                    coordIndexList.add(ci);
                    if (enableTex) {
                        texCoordIndexList.add(tci);
                    } else {
                        texCoordIndexList.add(null);
                    }
                }
            } catch (NumberFormatException e) {
                throw new InvalidObjException("invalid line (" + lineNum + "): " + src, e);
            }
        }
        while(src != null);
    }
}
