/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2008/03/15
 */
package net.cattaka.converttosculpt.logic;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;


import net.cattaka.converttosculpt.data.BVector3d;
import net.cattaka.converttosculpt.data.MeshData;
import net.cattaka.converttosculpt.exception.InvalidMeshDataException;
import net.cattaka.converttosculpt.exception.InvalidVrmlException;


public class MeshConverterVrml implements MeshConverter
{
    public void writeMeshData(MeshData meshData, OutputStream out) throws IOException,InvalidMeshDataException {
        String vrmlStr = convertMeshDataToVrml(meshData);
        
        OutputStreamWriter writer = new OutputStreamWriter(out);
        writer.write(vrmlStr);
        writer.flush();
    }

    
    /**
     * メッシュデータをVRML形式の文字列に変換する。
     * 
     * @param meshData
     * @return
     */
    public static String convertMeshDataToVrml(MeshData meshData)
    {
        //テキストデータ作成
        String name = meshData.getName();
        BVector3d[] coordinate = meshData.getCoordinate();
        BVector3d[] texCoordinate = meshData.getTexCoordinate();
        short[][] coordIndex = meshData.getCoordIndex();
        short[][] texCoordIndex = meshData.getTexCoordIndex();
        
        DecimalFormat df = new DecimalFormat();
        df.setMinimumFractionDigits(6);
        df.setMaximumFractionDigits(6);
        
        StringBuilder sb = new StringBuilder(1024*1024);
        String coordinateStr;
        String texCoordinateStr;
        String coordIndexStr;
        String texCoordIndexStr;
        
        //coordinate作成
        sb.delete(0, sb.length());
        for (int i=0;i<coordinate.length;i++) {
            if (i > 0) {
                sb.append(",\n");
            }
            BVector3d v = coordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
            sb.append(" ");
            sb.append(df.format(v.z));
        }
        coordinateStr = sb.toString();
    
        //coordIndex作成
        sb.delete(0, sb.length());
        for (int i=0;i<coordIndex.length;i++) {
            if (i > 0) {
                sb.append(",\n");
            }
            short[] ci = coordIndex[i];
            for (int j=0;j<ci.length;j++) {
                sb.append(ci[j]);
                sb.append(",");
            }
            sb.append("-1");
        }
        coordIndexStr = sb.toString();
    
        //texCoordinate作成
        sb.delete(0, sb.length());
        for (int i=0;i<texCoordinate.length;i++) {
            if (i > 0) {
                sb.append(",\n");
            }
            BVector3d v = texCoordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
        }
        texCoordinateStr = sb.toString();
        
        //texCoordIndex作成
        sb.delete(0, sb.length());
        for (int i=0;i<texCoordIndex.length;i++) {
            if (i > 0) {
                sb.append(",\n");
            }
            short[] ci = texCoordIndex[i];
            for (int j=0;j<ci.length;j++) {
                sb.append(ci[j]);
                sb.append(",");
            }
            sb.append("-1");
        }
        texCoordIndexStr = sb.toString();
    
        
        
        //テキストデータ作成
        sb.delete(0, sb.length());
    
        sb.append("#VRML V1.0 ascii\n");
        sb.append("# ConvertToSculpt build ");
        sb.append(VersionBundle.getBuildNumber());
        sb.append("\n");
    
        sb.append("Separator {\n");
        sb.append("Switch {\n");
        sb.append("\tDEF " + meshData.getName() + "\n");
        sb.append("\tSeparator {\n");
        sb.append("\t\tCoordinate3 {\n");
        sb.append("\t\t\tpoint [\n");
        sb.append(coordinateStr);
        sb.append("\t\t\t]\n");
        sb.append("\t\t}\n");
    
        sb.append("\t\tTextureCoordinate2 {\n");
        sb.append("\t\t\tpoint [\n");
        sb.append(texCoordinateStr);
        sb.append("\t\t\t]\n");
        sb.append("\t\t}\n");
        
        sb.append("\t\tIndexedFaceSet {\n");
        sb.append("\t\t\tcoordIndex [\n");
        sb.append(coordIndexStr);
        sb.append("\t\t\t]\n");
        sb.append("\t\t\ttextureCoordIndex [\n");
        sb.append(texCoordIndexStr);
        sb.append("\t\t\t]\n");
        sb.append("\t\t}\n");
    
        sb.append("\t}\n");
        sb.append("}\n");
    
        sb.append("Separator {\n");
        sb.append("\tSeparator {\n");
        sb.append("\t\tMatrixTransform {\n");
        sb.append("\t\tmatrix\n"); 
        sb.append("\t\t\t1.000000 0.000000 0.000000 0.000000\n");
        sb.append("\t\t\t0.000000 1.000000 0.000000 0.000000\n");
        sb.append("\t\t\t0.000000 0.000000 1.000000 0.000000\n");
        sb.append("\t\t\t0.000000 0.000000 0.000000 1.000000\n");
        sb.append("\t\t}\n");
        sb.append("\t\tUSE " + meshData.getName() + "\n");
        sb.append("\t}\n");
        sb.append("}\n");
        sb.append("}\n");
        
        return sb.toString();
    }

    /**
     * 
     * @param in       入力
     * @param result    出力
     * @param messages  出力
     * @throws InvalidX3dException
     * @throws IOException
     */
    public void readMeshData(InputStream in, List<MeshData> result, List<String> messages) throws IOException,InvalidMeshDataException
    {
        InputStreamReader reader = new InputStreamReader(in);

        StringBuilder sb = new StringBuilder(1024*1024);
        {
            int i;
            while((i=reader.read()) != -1) {
                sb.append((char)i);
            }
        }
        
        String src = sb.toString();
        
        int n = 0;
        ArrayList<String> defList = StringUtil.getElementFullString(src, "DEF", '{', '}');
        {
            for (int j=0;j<defList.size();j++) {
                String[] splited = defList.get(j).split("\\s+", 3);
                if (splited.length < 3) {
                    continue;
                }
                String name = splited[1];
                
                ArrayList<String> separatorList = StringUtil.getElementString(splited[2], "Separator", '{', '}');
                for (String str : separatorList) {
                    if (str.indexOf("Separator") != -1) {
                        continue;
                    }
                    try {
                        MeshData md = convertSeparatorElementToMeshData(name, str);
                        if (md != null) {
                            result.add(md);
                            n++;
                        }
                    } catch (Exception e) {
                        messages.add(e.getMessage());
                        e.printStackTrace();
                    }
                }
            }
        }
        
        if (n == 0) {
            throw new InvalidVrmlException("Load VRML failed.");
        }
    }

    /**
     * 
     * @param name
     * @param sepElement
     * @return                      success->meshdata, otherwise->null
     * @throws InvalidVrmlException
     * @throws IOException
     */
    public static MeshData convertSeparatorElementToMeshData(String name, String sepElement) throws InvalidVrmlException, IOException
    {
        MeshData meshData = new MeshData();
        meshData.setName(name);

        ArrayList<String> faceSetList = StringUtil.getElementString(sepElement, "IndexedFaceSet", '{', '}');
        if (faceSetList.size() > 0) {
            String faceSetElement = faceSetList.get(0);
            ArrayList<String> coordIndexList = StringUtil.getElementString(faceSetElement, "coordIndex", '[', ']');
            if (coordIndexList.size() > 0) {
                try {
                    meshData.setCoordIndex(convertStrToCoordIndex(coordIndexList.get(0)));
                } catch (NumberFormatException e) {
                    throw new InvalidVrmlException(name + " : coordIndex is invalid.", e);
                }
            }
            
            ArrayList<String> texCoordIndexList = StringUtil.getElementString(faceSetElement, "textureCoordIndex", '[', ']');
            if (texCoordIndexList.size() > 0) {
                try {
                    meshData.setTexCoordIndex(convertStrToCoordIndex(texCoordIndexList.get(0)));
                } catch (NumberFormatException e) {
                    throw new InvalidVrmlException(name + " : textureCoordIndex is invalid.", e);
                }
            }
        }

        ArrayList<String> coordinateList = StringUtil.getElementString(sepElement, "Coordinate3", '{', '}');
        if (coordinateList.size() > 0) {
            ArrayList<String> pointList = StringUtil.getElementString(coordinateList.get(0), "point", '[',']');
            if (pointList != null) {
                try {
                    meshData.setCoordinate(convertStrToCordinate(pointList.get(0)));
                } catch (NumberFormatException e) {
                    throw new InvalidVrmlException(name + " : Coordinate3 is invalid.", e);
                }
            }
        }
        
        ArrayList<String> texCoordinateList = StringUtil.getElementString(sepElement, "TextureCoordinate2", '{', '}');
        if (texCoordinateList.size() > 0) {
            ArrayList<String> pointList = StringUtil.getElementString(texCoordinateList.get(0), "point", '[',']');
            if (pointList != null) {
                try {
                    meshData.setTexCoordinate(convertStrToTexCordinate(pointList.get(0)));
                } catch (NumberFormatException e) {
                    throw new InvalidVrmlException(name + " : TextureCoordinate2 is invalid.", e);
                }
            }
        }
        
        if (meshData.getCoordinate() == null || meshData.getCoordIndex() == null) {
            return null;
        }
        
        return meshData;
    }

    
    /**
     * 文字列を分割し、CoordIndexを表すshort[][]に変換する。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static short[][] convertStrToCoordIndex(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
    
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t = tmp.split("(,\\s*)|\\s+");
        if (t.length == 1 && t[0].length() ==0) {
            return new short[0][];
        }
        
        short[] buf = new short[t.length];
        short[][] resultBuf = new short[t.length][];
        int p = 0;
        int c = 0;
        
        for (int i=0;i<t.length;i++) {
            short n = Short.parseShort(t[i].trim());
            if (n == -1) {
                resultBuf[p] = new short[c];
                System.arraycopy(buf, 0, resultBuf[p], 0, c);
                p++;
                c = 0;
            } else {
                buf[c] = n;
                c++;
            }
        }
        short[][] result = new short[p][];
        System.arraycopy(resultBuf, 0, result, 0, p);
        return result;
    }

    /**
     * 文字列を分割、パージングして３次元ベクトルの配列に変換する。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static BVector3d[] convertStrToCordinate(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
        
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t1 = tmp.split(",\\s*");
        if (t1.length == 1 && t1[0].length() == 0) {
            return new BVector3d[0];
        }
        BVector3d[] result = new BVector3d[t1.length];
        for (int i=0;i<t1.length;i++) {
            String[] t2 = t1[i].split("\\s+");
            if (t2.length != 3) {
                throw new RuntimeException("Invalid Coordinate : " + t1[i]);
            }
            double x = Double.parseDouble(t2[0].trim());
            double y = Double.parseDouble(t2[1].trim());
            double z = Double.parseDouble(t2[2].trim());
            result[i] = new BVector3d(x,y,z);
        }
        return result;
    }

    /**
     * 文字列を分割、パージングして２次元ベクトルの配列に変換する。
     * ３次元ベクトルを返すがx,y座標しか使用していない。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static BVector3d[] convertStrToTexCordinate(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
    
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t1 = tmp.split(",\\s*");
        if (t1.length == 1 && t1[0].length() == 0) {
            return new BVector3d[0];
        }
        BVector3d[] result = new BVector3d[t1.length];
        for (int i=0;i<t1.length;i++) {
            String[] t2 = t1[i].split("\\s+");
            if (t2.length != 2) {
                throw new RuntimeException("Invalid Coordinate : " + t1[i]);
            }
            double x = Double.parseDouble(t2[0].trim());
            double y = Double.parseDouble(t2[1].trim());
            result[i] = new BVector3d(x,y,0);
        }
        return result;
    }
}
