/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2008/03/15
 */
package net.cattaka.converttosculpt.logic;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DecimalFormat;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import net.cattaka.converttosculpt.data.BVector3d;
import net.cattaka.converttosculpt.data.MeshData;
import net.cattaka.converttosculpt.exception.InvalidMeshDataException;
import net.cattaka.converttosculpt.exception.InvalidX3dException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;


public class MeshConverterX3d implements MeshConverter
{
    public void writeMeshData(MeshData meshData, OutputStream out) throws IOException,InvalidMeshDataException {
        Document document = convertMeshDataToX3d(meshData);
        try
        {
            TransformerFactory transFactory = TransformerFactory.newInstance();
            Transformer transformer = transFactory.newTransformer();
            
            DOMSource source = new DOMSource(document);
            StreamResult result = new StreamResult(out); 
            transformer.transform(source, result);
            
        }catch(TransformerException e)
        {
            throw new IOException(e.toString());
        }
    }

    /**
     * メッシュデータをX3D形式のXMLに変換する
     * 
     * @return
     */
    public static Document convertMeshDataToX3d(MeshData meshData) {
        Document document = XMLConverter.createNewDocument("X3D");
        Element rootElement = document.getDocumentElement();
        rootElement.setAttribute("xsd:noNamespaceSchemaLocation", "http://www.web3d.org/specifications/x3d-3.0.xsd");
        rootElement.setAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema-instance");
        rootElement.setAttribute("profile", "Immersive");
        rootElement.setAttribute("version", "3.0");
    
        Element sceneElement = document.createElement("Scene");
        rootElement.appendChild(sceneElement);
        
        Element transformElement = createTransformElement(document, meshData);
        sceneElement.appendChild(transformElement);
        
        return document;
    }

    /**
     * 文字列を分割し、CoordIndexを表すshort[][]に変換する。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static short[][] convertStrToCoordIndex(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
    
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t = tmp.split("(,\\s*)|\\s+");
        if (t.length == 1 && t[0].length() ==0) {
            return new short[0][];
        }
        
        short[] buf = new short[t.length];
        short[][] resultBuf = new short[t.length][];
        int p = 0;
        int c = 0;
        for (int i=0;i<t.length;i++) {
            try {
                short n = Short.parseShort(t[i].trim());
                if (n == -1) {
                    resultBuf[p] = new short[c];
                    System.arraycopy(buf, 0, resultBuf[p], 0, c);
                    p++;
                    c = 0;
                } else {
                    buf[c] = n;
                    c++;
                }
            } catch (RuntimeException e) {
                throw e;
            }
        }
        short[][] result = new short[p][];
        System.arraycopy(resultBuf, 0, result, 0, p);
        return result;
    }

    /**
     * 文字列を分割、パージングして３次元ベクトルの配列に変換する。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static BVector3d[] convertStrToCordinate(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
        
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t1 = tmp.split(",\\s*");
        if (t1.length == 1 && t1[0].length() == 0) {
            return new BVector3d[0];
        }
        BVector3d[] result = new BVector3d[t1.length];
        for (int i=0;i<t1.length;i++) {
            String[] t2 = t1[i].split("\\s+");
            if (t2.length != 3) {
                throw new RuntimeException("Invalid Coordinate : " + t1[i]);
            }
            double x = Double.parseDouble(t2[0].trim());
            double y = Double.parseDouble(t2[1].trim());
            double z = Double.parseDouble(t2[2].trim());
            result[i] = new BVector3d(x,y,z);
        }
        return result;
    }

    /**
     * 文字列を分割、パージングして２次元ベクトルの配列に変換する。
     * ３次元ベクトルを返すがx,y座標しか使用していない。
     * 
     * @param arg
     * @return
     * @throws NumberFormatException 数値をカンマで区切った文字列以外が与えられた場合
     */
    private static BVector3d[] convertStrToTexCordinate(String arg) throws NumberFormatException {
        if (arg == null) {
            throw new NullPointerException();
        }
    
        String tmp = StringUtil.stripChars(arg.trim(), "", ",");
        String[] t1 = tmp.split(",\\s*");
        if (t1.length == 1 && t1[0].length() == 0) {
            return new BVector3d[0];
        }
        BVector3d[] result = new BVector3d[t1.length];
        for (int i=0;i<t1.length;i++) {
            String[] t2 = t1[i].split("\\s+");
            if (t2.length != 2) {
                throw new RuntimeException("Invalid Coordinate : " + t1[i]);
            }
            double x = Double.parseDouble(t2[0].trim());
            double y = Double.parseDouble(t2[1].trim());
            result[i] = new BVector3d(x,y,0);
        }
        return result;
    }

    /**
     * 
     * @param in       入力
     * @param result    出力
     * @param messages  出力
     * @throws InvalidX3dException
     * @throws IOException
     */
    public void readMeshData(InputStream in, List<MeshData> result, List<String> messages) throws IOException,InvalidMeshDataException
    {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance(); 
        Document document;
        try {
            document = factory.newDocumentBuilder().parse(in);
        } catch (SAXException e) {
            throw new InvalidX3dException(e);
        } catch (ParserConfigurationException e) {
            throw new InvalidX3dException(e);
        }
        
        Element root = document.getDocumentElement();
        NodeList tes = root.getElementsByTagName("Transform");
        for (int i=0;i<tes.getLength();i++) {
            Element te = (Element)tes.item(i);
            try {
                MeshData md = convertTransformElementToMeshData(te);
                result.add(md);
            } catch (Exception e) {
                e.printStackTrace();
                messages.add(e.getMessage());
            }
        }
    }

    /**
     * 
     * @param element
     * @throws InvalidX3dException
     * @throws IOException
     */
    private static MeshData convertTransformElementToMeshData(Element transformElement) throws InvalidX3dException, IOException
    {
        MeshData meshData = new MeshData();
        if (!(transformElement.getNodeName().equals("Transform"))) {
            throw new InvalidX3dException("This is not Transform element");
        }
        
        String name = transformElement.getAttribute("DEF");
        meshData.setName(name);
        Element shapeElement = (Element)transformElement.getElementsByTagName("Shape").item(0);
        if (shapeElement == null) {
            throw new InvalidX3dException(name + " : Shape element is not found.");
        }
        
        Element faceSetElement = (Element)shapeElement.getElementsByTagName("IndexedFaceSet").item(0);
        if (faceSetElement != null) {
            String coordIndexStr = faceSetElement.getAttribute("coordIndex");
            String texCoordIndexStr = faceSetElement.getAttribute("texCoordIndex");
            if (coordIndexStr != null && coordIndexStr.length() > 0) {
                try {
                    meshData.setCoordIndex(convertStrToCoordIndex(coordIndexStr));
                } catch (NumberFormatException e) {
                    throw new InvalidX3dException(name + " : coordIndex is invalid.", e);
                }
            }
            if (texCoordIndexStr != null && texCoordIndexStr.length() > 0) {
                try {
                    meshData.setTexCoordIndex(convertStrToCoordIndex(texCoordIndexStr));
                } catch (NumberFormatException e) {
                    throw new InvalidX3dException(name + " : texCoordIndex is invalid.", e);
                }
            }
        }
        
        Element coordinateElement = (Element)faceSetElement.getElementsByTagName("Coordinate").item(0);
        if (coordinateElement != null) {
            String coordinateStr = coordinateElement.getAttribute("point");
            if (coordinateStr != null && coordinateStr.length() > 0) {
                try {
                    meshData.setCoordinate(convertStrToCordinate(coordinateStr));
                } catch (NumberFormatException e) {
                    throw new InvalidX3dException(name + " : point attribute of Coordinate is invalid.");
                }
            }
        }
        
        Element texCoordinateElement = (Element)faceSetElement.getElementsByTagName("TextureCoordinate").item(0);
        if (texCoordinateElement != null) {
            String texCoordinateStr = texCoordinateElement.getAttribute("point");
            if (texCoordinateStr != null && texCoordinateStr.length() > 0) {
                try {
                    meshData.setTexCoordinate(convertStrToTexCordinate(texCoordinateStr));
                } catch (NumberFormatException e) {
                    throw new InvalidX3dException(name + " : point attribute of TextureCoordinate is invalid.");
                }
            }
        }
        return meshData;
    }

    /**
     * 
     * @param document
     * @param meshData
     * @return
     */
    private static Element createTransformElement(Document document, MeshData meshData) {
        //テキストデータ作成
        String name = meshData.getName();
        BVector3d[] coordinate = meshData.getCoordinate();
        BVector3d[] texCoordinate = meshData.getTexCoordinate();
        short[][] coordIndex = meshData.getCoordIndex();
        short[][] texCoordIndex = meshData.getTexCoordIndex();
        
        DecimalFormat df = new DecimalFormat();
        df.setMinimumFractionDigits(6);
        df.setMaximumFractionDigits(6);
        
        StringBuilder sb = new StringBuilder(1024*1024);
        String coordinateStr;
        String texCoordinateStr;
        String coordIndexStr;
        String texCoordIndexStr;
        
        //coordinate作成
        sb.delete(0, sb.length());
        for (int i=0;i<coordinate.length;i++) {
            if (i > 0) {
                sb.append(",");
            }
            BVector3d v = coordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
            sb.append(" ");
            sb.append(df.format(v.z));
        }
        coordinateStr = sb.toString();
    
        //coordIndex作成
        sb.delete(0, sb.length());
        for (int i=0;i<coordIndex.length;i++) {
            if (i > 0) {
                sb.append(",");
            }
            short[] ci = coordIndex[i];
            for (int j=0;j<ci.length;j++) {
                sb.append(ci[j]);
                sb.append(" ");
            }
            sb.append("-1");
        }
        coordIndexStr = sb.toString();
    
        //texCoordinate作成
        sb.delete(0, sb.length());
        for (int i=0;i<texCoordinate.length;i++) {
            if (i > 0) {
                sb.append(",");
            }
            BVector3d v = texCoordinate[i];
            sb.append(df.format(v.x));
            sb.append(" ");
            sb.append(df.format(v.y));
        }
        texCoordinateStr = sb.toString();
        
        //texCoordIndex作成
        sb.delete(0, sb.length());
        for (int i=0;i<texCoordIndex.length;i++) {
            if (i > 0) {
                sb.append(",");
            }
            short[] ci = texCoordIndex[i];
            for (int j=0;j<ci.length;j++) {
                sb.append(ci[j]);
                sb.append(",");
            }
            sb.append("-1");
        }
        texCoordIndexStr = sb.toString();
        
        //XMLドキュメント作成
        Element transformElement = document.createElement("Transform");
        Element shapeElement = document.createElement("Shape");
        Element indexedFaceSetElement = document.createElement("IndexedFaceSet");
        Element coordinateElement = document.createElement("Coordinate");
        Element texCoordinateElement = document.createElement("TextureCoordinate");
    
        coordinateElement.setAttribute("point", coordinateStr);
        texCoordinateElement.setAttribute("point", texCoordinateStr);
        indexedFaceSetElement.setAttribute("coordIndex", coordIndexStr);
        indexedFaceSetElement.setAttribute("texCoordIndex", texCoordIndexStr);
        
        //document.getDocumentElement().appendChild(transformElement);
        transformElement.appendChild(shapeElement);
        shapeElement.appendChild(indexedFaceSetElement);
        indexedFaceSetElement.appendChild(coordinateElement);
        indexedFaceSetElement.appendChild(texCoordinateElement);
        
        transformElement.setAttribute("DEF", name);
        transformElement.setAttribute("translation", "0.0 0.0 0.0");
        indexedFaceSetElement.setAttribute("solid", "true");
        coordinateElement.setAttribute("DEF", "coord_" + name);
        
        return transformElement;
    }

}
