/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: Feb 16, 2008
 */
package net.cattaka.converttosculpt.logic;

import java.awt.image.BufferedImage;

import net.cattaka.converttosculpt.data.BVector3d;
import net.cattaka.converttosculpt.data.MeshData;
import net.cattaka.converttosculpt.data.SizeInfo;
import net.cattaka.converttosculpt.exception.InvalidArgumentException;
import net.cattaka.converttosculpt.exception.InvalidMeshDataException;

public abstract class MeshUtil
{
    /**
     * 与えられたメッシュデータのサイズ情報を算出する。
     * @param meshData 入力メッシュデータ
     * @return サイズ情報
     */
    public static SizeInfo calcSizeInfo(MeshData meshData) {
        SizeInfo sizeInfo = new SizeInfo();

        BVector3d[] src = meshData.getCoordinate();
        if (src.length > 0) {
            BVector3d v = src[0];
            sizeInfo.setMinX(v.x);
            sizeInfo.setMaxX(v.x);
            sizeInfo.setMinY(v.y);
            sizeInfo.setMaxY(v.y);
            sizeInfo.setMinZ(v.z);
            sizeInfo.setMaxZ(v.z);
            for (int i=1;i<src.length;i++) {
                v = src[i];
                
                if (v.x < sizeInfo.getMinX()) {
                    sizeInfo.setMinX(v.x);
                }
                if (v.y < sizeInfo.getMinY()) {
                    sizeInfo.setMinY(v.y);
                }
                if (v.z < sizeInfo.getMinZ()) {
                    sizeInfo.setMinZ(v.z);
                }
                if (v.x > sizeInfo.getMaxX()) {
                    sizeInfo.setMaxX(v.x);
                }
                if (v.y > sizeInfo.getMaxY()) {
                    sizeInfo.setMaxY(v.y);
                }
                if (v.z > sizeInfo.getMaxZ()) {
                    sizeInfo.setMaxZ(v.z);
                }
            }
        }
        return sizeInfo;
    }
    
    /**
     * 与えたサイズ情報に会わせてメッシュデータを正規化する。
     * サイズ情報の値によっては-0.5〜0.5のレンジに収まらない場合がある。
     * @param meshData
     * @param sizeInfo
     */
    public static void normalizeCoordinate(MeshData meshData, SizeInfo sizeInfo) {
        BVector3d range = new BVector3d();
        BVector3d center = new BVector3d();
        center.x = (sizeInfo.getMaxX() + sizeInfo.getMinX()) / 2.0;
        center.y = (sizeInfo.getMaxY() + sizeInfo.getMinY()) / 2.0;
        center.z = (sizeInfo.getMaxZ() + sizeInfo.getMinZ()) / 2.0;
        
        range.x = (sizeInfo.getMaxX() - sizeInfo.getMinX());
        range.y = (sizeInfo.getMaxY() - sizeInfo.getMinY());
        range.z = (sizeInfo.getMaxZ() - sizeInfo.getMinZ());

        if (range.x == 0) {
            range.x = 1;
        }
        if (range.y == 0) {
            range.y = 1;
        }
        if (range.z == 0) {
            range.z = 1;
        }
        
        BVector3d[] src = meshData.getCoordinate();
        for (int i=0;i<src.length;i++) {
            src[i].sub(center);
            src[i].x = src[i].x / range.x;
            src[i].y = src[i].y / range.y;
            src[i].z = src[i].z / range.z;
        }
    }
    /**
     * 自動的にメッシュデータを-0.5〜0.5のレンジに正規化する。
     * @param meshData
     */
    public static void autoNormalizeCoordinate(MeshData meshData) {
        SizeInfo sizeInfo = calcSizeInfo(meshData);
        normalizeCoordinate(meshData, sizeInfo);
    }
    
    /**
     * 与えられたメッシュデータをスカルプテクスチャに変換して出力する。
     * 頂点座標が-0.5〜0.5のレンジに収まらない場合は上限値（下限値）に割り当てられる。
     * @param meshData  入力メッシュデータ
     * @param dist      出力先
     * @param numX      頂点数(X)
     * @param numY      頂点数(Y)
     * @throws InvalidMeshDataException
     */
    
    public static void drawMesh(MeshData meshData, BufferedImage dist, int numX, int numY) throws InvalidMeshDataException {
        if (dist.getWidth() != numX) {
            throw new InvalidArgumentException("Image width must be equal to the number of vertices+1.");
        }
        if (dist.getHeight() != numY) {
            throw new InvalidArgumentException("Image width must be equal to the number of vertices+1.");
        }
        meshData.isValid();
        
        BVector3d[] vs = meshData.getCoordinate();
        
        // draw Image
        BVector3d[] texCoordinates = meshData.getTexCoordinate();
        short[][] ci = meshData.getCoordIndex();
        short[][] tci = meshData.getTexCoordIndex();
        outer : for (int i=0;i<tci.length;i++) {
            int[] xs;
            int[] ys;
            BVector3d[] cs;
            short[] t = tci[i];
            if (t == null || t.length < 3 || 4 < t.length) {
                //三角形と四角形以外は諦める
                continue;
            }
            xs = new int[t.length];
            ys = new int[t.length];
            cs = new BVector3d[t.length];
            for (int k=0;k<t.length;k++) {
                BVector3d target = texCoordinates[t[k]];
                if (target == null) {
                    continue outer;
                }
                xs[k] = (int)Math.round((numX-1) * target.x);
                ys[k] = (int)Math.round((numY-1) * target.y);
                cs[k] = vs[ci[i][k]];
            }
            //System.out.println(texCoordinates[t[0]].x);
            PolygonDrower.drawPorygon(dist, xs, ys, cs);
        }
    }

    /**
     * BVector3dからMeshDataを作成する。
     * メッシュはnx*nyの頂点を持つ長方形を変形した物とする。
     * 
     * @param name          名称
     * @param vertices      頂点の配列の配列
     * @param nx            x方向の頂点数
     * @param ny            y方向の頂点数
     * @param closeSideEdge 側面を閉じるオプション 
     * @return              メッシュデータ
     */
    public static MeshData convertBVector3dToMeshData(String name, BVector3d[][] vertices, int nx, int ny, boolean closeSideEdge) {
        if (closeSideEdge) {
            BVector3d[] vs = new BVector3d[nx*ny];
            for (int j=0;j<ny;j++) {
                for (int i=0;i<nx;i++) {
                    vs[nx*j + i] = vertices[i][j];
                }
            }
    
            BVector3d[] tvs = new BVector3d[(nx+1)*ny];
            for (int j=0;j<ny;j++) {
                for (int i=0;i<nx+1;i++) {
                    double tx = ((double)i/(double)(nx)); 
                    double ty = ((double)j/(double)(ny-1)); 
                    tvs[(nx+1)*j + i] = new BVector3d(tx, ty, 0);
                }            
            }
            
            short[][] coordIndex;
            short[][] texCoordIndex;
            coordIndex = new short[(nx)*(ny-1)][];
            texCoordIndex = new short[(nx)*(ny-1)][];
            
            for (int j=0;j<ny-1;j++) {
                for (int i=0;i<nx;i++) {
                    int cidx = nx * j + i;
                    if (i == nx-1) {
                        coordIndex[cidx] = new short[]{
                                (short)(nx*j + i),
                                (short)(nx*j),
                                (short)(nx*(j+1)),
                                (short)(nx*(j+1) + i)
                            };
                    } else {
                        coordIndex[cidx] = new short[]{
                                (short)(nx*j + i),
                                (short)(nx*j + i     + 1),
                                (short)(nx*(j+1) + i + 1),
                                (short)(nx*(j+1) + i)
                            };
                    }
                }
            }
            for (int j=0;j<ny-1;j++) {
                for (int i=0;i<nx;i++) {
                    int cidx = nx * j + i;
                    texCoordIndex[cidx] = new short[]{
                            (short)((nx+1)*j + i),
                            (short)((nx+1)*j + i     + 1),
                            (short)((nx+1)*(j+1) + i + 1),
                            (short)((nx+1)*(j+1) + i)
                        };
                }
            }
            
            return new MeshData(name, coordIndex, vs, texCoordIndex, tvs);
        } else {
            BVector3d[] vs = new BVector3d[nx*ny];
            for (int j=0;j<ny;j++) {
                for (int i=0;i<nx;i++) {
                    vs[nx*j + i] = vertices[i][j];
                }
            }
    
            BVector3d[] tvs = new BVector3d[nx*ny];
            for (int j=0;j<ny;j++) {
                for (int i=0;i<nx;i++) {
                    double tx = ((double)i/(double)(nx-1)); 
                    double ty = ((double)j/(double)(ny-1)); 
                    tvs[nx*j + i] = new BVector3d(tx, ty, 0);
                }            
            }
            
            short[][] coordIndex;
            short[][] texCoordIndex;
            int coordIndexNum = (nx-1)*(ny-1);
            coordIndex = new short[coordIndexNum][];
            texCoordIndex = new short[coordIndexNum][];
            
            //側面を閉じない場合
            for (int i=0;i<coordIndexNum;i++) {
                int s = i / (nx-1);
                short c1 = (short)(i+s);
                short c2 = (short)(i+s+1);
                short c3 = (short)(i+s+1+nx);
                short c4 = (short)(i+s  +nx);
                coordIndex[i] = new short[]{c1,c2,c3,c4};
                texCoordIndex[i] = new short[]{c1,c2,c3,c4};
            }
            
            return new MeshData(name, coordIndex, vs, texCoordIndex, tvs);
        }
    }
}
