/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2008/02/17
 */
package net.cattaka.converttosculpt.logic;

import java.awt.Color;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;

import net.cattaka.converttosculpt.data.BVector2d;
import net.cattaka.converttosculpt.data.BVector3d;



/**
 * 複数の色のポリゴンをレンダリングするためのロジック。
 * 
 * @author cattaka
 */
public class PolygonDrower
{
    /**
     * 与えられた多角形情報と色情報を用いて、頂点間の色を補間しながら塗りつぶす。
     * 頂点数は3と4のみをサポート。
     * 
     * @param dist 出力先
     * @param xs 頂点のx座標の配列
     * @param ys 頂点のy座標の配列
     * @param cs 頂点の色を表すベクトルの配列
     */
    public static void drawPorygon(BufferedImage dist, int[] xs, int[] ys, BVector3d[] cs) {
        if (xs.length != ys.length || ys.length != cs.length) {
            throw new RuntimeException("Number of indecies is mismatch.");
        }
        if (xs.length == 3) {
            drawTriangle(dist, xs[0], ys[0], cs[0], xs[1], ys[1], cs[1], xs[2], ys[2], cs[2]);
        } else if (xs.length == 4) {
            drawTriangle(dist, xs[0], ys[0], cs[0], xs[1], ys[1], cs[1], xs[2], ys[2], cs[2]);
            drawTriangle(dist, xs[2], ys[2], cs[2], xs[3], ys[3], cs[3], xs[0], ys[0], cs[0]);
        } else {
            throw new RuntimeException("Number of indecies is invalid. 3 or 4 are supported.");
        }
    }
    
    /**
     * 与えられた多角形情報と色情報を用いて、頂点間の色を補間しながら三角形を塗りつぶす。
     * 
     * @param dist 出力先
     * @param x1 頂点1のx座標
     * @param y1 頂点1のy座標
     * @param c1 頂点1の色
     * @param x2 頂点2のx座標
     * @param y2 頂点2のy座標
     * @param c2 頂点2の色
     * @param x3 頂点3のx座標
     * @param y3 頂点3のy座標
     * @param c3 頂点3の色
     */
    public static void drawTriangle(BufferedImage dist, int x1, int y1, BVector3d c1, int x2, int y2, BVector3d c2, int x3, int y3, BVector3d c3) {
        ExPolygon pg = new ExPolygon(new int[]{x1,x2,x3}, new int[]{y1,y2,y3}, 3);
        Rectangle bound = pg.getBounds();
        for (int i=bound.x;i<=bound.x+bound.width;i++) {
            for (int j=bound.y;j<=bound.y+bound.height;j++) {
                if (j<0 || dist.getHeight() <= j || i<0 || dist.getWidth() <= i) {
                    continue;
                }
                Color c;
                if (i==x1 && j==y1) {
                    c = convertVector3dToColor(c1);
                    dist.setRGB(i, j, c.getRGB());
                } else if (i==x2 && j==y2) {
                    c = convertVector3dToColor(c2);
                    dist.setRGB(i, j, c.getRGB());
                } else if (i==x3 && j==y3) {
                    c = convertVector3dToColor(c3);
                    dist.setRGB(i, j, c.getRGB());
                } else {
                    if (pg.containsEx(i,j)) {
                        BVector2d g = new BVector2d(i,j);
                        BVector2d p1 = new BVector2d(x1,y1);
                        BVector2d p2 = new BVector2d(x2,y2);
                        BVector2d p3 = new BVector2d(x3,y3);
                        
                        c = convertVector3dToColor(calcColor(g, p1, c1, p2, c2, p3, c3));
                        dist.setRGB(i, j, c.getRGB());
                    }
                }
            }
        }
    }

    /**
     * 三つの頂点に囲まれた面上の1つの点の色を計算する。
     * @param g     対象となる点の座標
     * @param p1    頂点1の座標
     * @param c1    頂点1の色
     * @param p2    頂点2の座標
     * @param c2    頂点2の色
     * @param p3    頂点3の座標
     * @param c3    頂点3の色
     * @return      対象となる点の色
     */
    private static BVector3d calcColor(BVector2d g, BVector2d p1, BVector3d c1, BVector2d p2, BVector3d c2, BVector2d p3, BVector3d c3) {
        BVector2d gmp1 = new BVector2d(g);
        gmp1.sub(p1);
        BVector2d p3mp1 = new BVector2d(p3);
        p3mp1.sub(p1);
        BVector2d p2mp3 = new BVector2d(p2);
        p2mp3.sub(p3);
        
        double a = BVector2d.getCross(gmp1, p2mp3);
        if (a != 0) {
            a = - BVector2d.getCross(gmp1, p3mp1) / a;
        } else {
            a = 0.5f;
        }
        
        BVector2d p23 = new BVector2d();
        BVector2d t = new BVector2d();
        t.set(p2);
        t.scale(a);
        p23.add(t);
        t.set(p3);
        t.scale(1.0 - a);
        p23.add(t);
        
        BVector2d p23mp1 = new BVector2d(p23);
        p23mp1.sub(p1);
        
              
        double b = p23mp1.getLength();
        if (b != 0) {
            b = 1.0 - gmp1.getLength() / b;
        } else {
            b = 1.0 / 3.0;
        }
        
        BVector3d c23 = new BVector3d();
        c23.scaleAdd(a, c2);
        c23.scaleAdd(1.0-a, c3);

        BVector3d result = new BVector3d();
        result.scaleAdd(b, c1);
        result.scaleAdd(1.0-b, c23);

        return result;
    }
    /**
     * 座標ベクトルに対応する色を返す。
     * 計算式は次の通り。<br>
     * 
     * r = v.x(-0.5 ~ 0.5)<br>
     * g = v.y(-0.5 ~ 0.5)<br>
     * b = v.z(-0.5 ~ 0.5)<br>
     * @param v 座標ベクトル
     * @return  ベクトルに対応する色
     */
    private static Color convertVector3dToColor(BVector3d v) {
        double r = (v.x + 0.5);
        double g = (v.y + 0.5);
        double b = (v.z + 0.5);
        
        if (r < 0.0) {
            r = 0.0;
        } else if (r > 1.0) {
            r = 1.0;
        }
        if (g < 0.0) {
            g = 0.0;
        } else if (g > 1.0) {
            g = 1.0;
        }
        if (b < 0.0) {
            b = 0.0;
        } else if (b > 1.0) {
            b = 1.0;
        }
        return new Color((float)r, (float)g, (float)b);
    }

}


/**
 * 標準のPolygonでは境界上の点はcontainsでfalseになる場合があるため、
 * 手を加えたバージョン。
 * 
 * @author cattaka
 */
class ExPolygon extends Polygon
{
    public ExPolygon(int xpoints[], int ypoints[], int npoints) {
        super(xpoints, ypoints, npoints);
    }
    
    public boolean containsEx(int x, int y)
    {
        return containsEx((double) x, (double) y);
    }

    /**
     * 境界上の点もtrueを返すように改造したもの
     * 
     * @param x
     * @param y
     * @return
     */
    public boolean containsEx(double x, double y)
    {
        if (npoints <= 2)
        {
            return false;
        }
        int hits = 0;

        int lastx = xpoints[npoints - 1];
        int lasty = ypoints[npoints - 1];
        int curx, cury;

        // Walk the edges of the polygon
        for (int i = 0; i < npoints; lastx = curx, lasty = cury, i++)
        {
            curx = xpoints[i];
            cury = ypoints[i];

            if (cury == lasty)
            {
                continue;
            }

            int leftx;
            if (curx < lastx)
            {
                if (x > lastx)
                {
                    continue;
                }
                leftx = curx;
            }
            else
            {
                if (x > curx)
                {
                    continue;
                }
                leftx = lastx;
            }

            double test1, test2;
            if (cury < lasty)
            {
                if (y < cury || y > lasty)
                {
                    continue;
                }
                if (x < leftx)
                {
                    hits++;
                    continue;
                }
                test1 = x - curx;
                test2 = y - cury;
            }
            else
            {
                if (y < lasty || y > cury)
                {
                    continue;
                }
                if (x < leftx)
                {
                    hits++;
                    continue;
                }
                test1 = x - lastx;
                test2 = y - lasty;
            }

            double t = (test2 / (lasty - cury) * (lastx - curx));
            if (test1 < t)
            {
                hits++;
            } else if (test1 == t) {
                return true;
            }
        }

        return ((hits & 1) != 0);
    }
}
