/*
 * Copyright 2008 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: Feb 16, 2008
 */
package net.cattaka.converttosculpt.ui;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import net.cattaka.converttosculpt.data.BMath;
import net.cattaka.converttosculpt.data.BVector3d;
import net.cattaka.converttosculpt.data.MeshData;
import net.cattaka.converttosculpt.data.SizeInfo;
import net.cattaka.converttosculpt.datainputpanel.DIPInfo;
import net.cattaka.converttosculpt.datainputpanel.DIPInfoInteger;
import net.cattaka.converttosculpt.datainputpanel.DataInputPanel;
import net.cattaka.converttosculpt.exception.InvalidArgumentException;
import net.cattaka.converttosculpt.exception.InvalidMeshDataException;
import net.cattaka.converttosculpt.exception.InvalidX3dException;
import net.cattaka.converttosculpt.logic.MeshConverter;
import net.cattaka.converttosculpt.logic.MeshUtil;
import net.cattaka.converttosculpt.logic.MeshConverterObj;
import net.cattaka.converttosculpt.logic.MeshConverterVrml;
import net.cattaka.converttosculpt.logic.MeshConverterX3d;
import net.cattaka.converttosculpt.logic.RenderingConf;

public class ConvertToSculptPanel extends JPanel implements ActionListener
{
    public static final int GRID_SIZE = 32;
    
    private JFileChooser fileChooser;

    private RenderingConf renderingConf;
    private MeshData meshData;
    private List<MeshData> meshDataList;

    private DataInputPanel dipRenderingConf;
    private FileSelectPanel fileSelectPanel;
    private JList listShapes;
    private InputSizeInfoPanel inputSizeInfoPanel;
    private DisplaySizeInfoPanel displaySizeInfoPanel;
    
    private HashMap<String, MeshConverter> meshConverterMap = new HashMap<String, MeshConverter>();
    
    public ConvertToSculptPanel(JFileChooser fileChooser)
    {
        this.meshConverterMap.put("x3d", new MeshConverterX3d());
        this.meshConverterMap.put("vrml", new MeshConverterVrml());
        this.meshConverterMap.put("obj", new MeshConverterObj());
        
        this.fileChooser = fileChooser;
        initialize();
    }
    
    private void initialize() {
        this.renderingConf = new RenderingConf();
        this.meshDataList = new ArrayList<MeshData>();
        
        //Create input panel
        DIPInfo[] dipInfo = new DIPInfo[4];
        dipInfo[0] = new DIPInfoInteger("Number of X vertex", String.valueOf(renderingConf.getVertexNumX()));
        dipInfo[1] = new DIPInfoInteger("Number of Y vertex", String.valueOf(renderingConf.getVertexNumY()));
        dipInfo[2] = new DIPInfoInteger("Image width", String.valueOf(renderingConf.getImageWidth()));
        dipInfo[3] = new DIPInfoInteger("Image height", String.valueOf(renderingConf.getImageHeight()));
        this.dipRenderingConf = new DataInputPanel(dipInfo);
        this.dipRenderingConf.setShowButtons(false);

        //Create panels
        this.fileSelectPanel = new FileSelectPanel();
        this.displaySizeInfoPanel = new DisplaySizeInfoPanel();
        this.inputSizeInfoPanel = new InputSizeInfoPanel();
        this.listShapes = new JList();
        this.listShapes.setVisibleRowCount(3);
        this.listShapes.addListSelectionListener(new ListSelectionListener() {
            public void valueChanged(ListSelectionEvent e)
            {
                ConvertToSculptPanel.this.updateListItem(ConvertToSculptPanel.this.listShapes.getSelectedIndex());
            }
        });
        JScrollPane panelListShapes = new JScrollPane(listShapes);
        
        //Create sparators
        JSeparator separator1 = new JSeparator();
        JSeparator separator2 = new JSeparator();
        JSeparator separator3 = new JSeparator();
        separator1.setPreferredSize(new Dimension(1,10));
        separator2.setPreferredSize(new Dimension(1,10));
        separator3.setPreferredSize(new Dimension(1,10));

        //create labels
        JLabel labelShapeList = new JLabel("Shapes list");
        JLabel labelConvertOption = new JLabel("Convert option");
        JLabel labelSculptedTexture = new JLabel("Sculpted Texture");

        //create convert panel
        JTabbedPane convertPanel = new JTabbedPane();
        {
            StandardConvertPanel standardConvertPanel = new StandardConvertPanel();
            ForceConvertPanel forceConvertPanel = new ForceConvertPanel();
            convertPanel.addTab("StandartConvert", standardConvertPanel);
            convertPanel.addTab("ForceConvert", forceConvertPanel);
        }
        
        
        //Make layout
        GridBagLayout gbl = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.weightx = 1.0;
        gbc.weighty = 0.0;
        gbc.gridx = 0;
        gbc.gridy = 0;
        
        gbc.gridwidth = 2;
        gbl.setConstraints(this.fileSelectPanel, gbc);
        gbc.gridy++;

        gbl.setConstraints(separator1, gbc);
        gbc.gridy++;
        gbl.setConstraints(labelShapeList, gbc);
        gbc.gridy++;
        gbc.gridwidth = 1;
        gbl.setConstraints(panelListShapes, gbc);
        gbc.gridx++;
        gbl.setConstraints(this.displaySizeInfoPanel, gbc);
        gbc.gridx--;
        gbc.gridy++;

        
        gbc.gridwidth = 2;
        gbl.setConstraints(separator2, gbc);
        gbc.gridy++;
        gbl.setConstraints(labelConvertOption, gbc);
        gbc.gridy++;
        gbc.gridwidth = 1;
        gbl.setConstraints(dipRenderingConf, gbc);
        gbc.gridx++;
        gbl.setConstraints(this.inputSizeInfoPanel, gbc);
        //gbl.setConstraints(dipSizeInfo, gbc);
        gbc.gridx--;
        gbc.gridy++;

        
        gbc.gridwidth = 2;
        gbl.setConstraints(separator3, gbc);
        gbc.gridy++;
        gbl.setConstraints(labelSculptedTexture, gbc);
        gbc.gridy++;
        gbc.fill = GridBagConstraints.BOTH;
        gbc.weightx = 1.0;
        gbc.weighty = 1.0;
        gbl.setConstraints(convertPanel, gbc);

        setLayout(gbl);
        add(fileSelectPanel);

        add(separator1);
        add(labelShapeList);
        add(panelListShapes);
        add(displaySizeInfoPanel);

        add(separator2);
        add(labelConvertOption);
        add(dipRenderingConf);
        add(inputSizeInfoPanel);

        add(separator3);
        add(labelSculptedTexture);

        add(convertPanel);
        
        //((DIPInfoFile)dipInfo[0]).setValue(new File("/home/cattaka/project/eclipse/ConvertToSculpt/plane_test02.x3d"));
    }
    

    public void actionPerformed(ActionEvent e)
    {
        try {
            if (e.getActionCommand().equals("load x3d")) {
                if (this.fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                    File inputFIle = this.fileChooser.getSelectedFile();
                    loadFile(inputFIle, "x3d");
                }
            } else if (e.getActionCommand().equals("load vrml")) {
                if (this.fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                    File inputFIle = this.fileChooser.getSelectedFile();
                    loadFile(inputFIle, "vrml");
                }
            } else if (e.getActionCommand().equals("load obj")) {
                if (this.fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                    File inputFIle = this.fileChooser.getSelectedFile();
                    loadFile(inputFIle, "obj");
                }
            }
        } catch (Exception ex) {
            JOptionPane.showMessageDialog(this, ex.getMessage());
            ex.printStackTrace();
        }
    }
    
    private void loadFile(File file, String type) throws IOException, InvalidX3dException {
        BufferedInputStream in = new BufferedInputStream(new FileInputStream(file),1024*1024);
        
        try {
            this.meshDataList.clear();
            ArrayList<String> messages = new ArrayList<String>();
            
            MeshConverter meshConverter = this.meshConverterMap.get(type);
            if (meshConverter == null) {
                throw new InvalidArgumentException("invalid type");
            }
            meshConverter.readMeshData(in, this.meshDataList, messages);
            
            for (String msg : messages) {
                JOptionPane.showMessageDialog(this, msg);
            }
            
            this.fileSelectPanel.setFileName(file.toString());
            ArrayList<String> listitems = new ArrayList<String>();
            for (MeshData md : this.meshDataList) {
                String name = md.getName();
                if (name == null || name.length() == 0) {
                    name = "Unnamed";
                }
                listitems.add(name);
            }
            this.listShapes.setListData(listitems.toArray());
            this.listShapes.setVisibleRowCount(3);
            if (this.meshDataList.size() > 0) {
                this.listShapes.setSelectedIndex(0);
                updateListItem(0);
            }
        } catch (IOException e) {
            this.fileSelectPanel.setFileName("");
            this.listShapes.setListData(new Object[0]);
            throw e;
        } catch (InvalidMeshDataException e) {
            this.fileSelectPanel.setFileName("");
            this.listShapes.setListData(new Object[0]);
            throw e;
        } finally {
            in.close();
        }
    }
    private void updateListItem(int selectedItem) {
        if (0 <= selectedItem && selectedItem < this.meshDataList.size()) {
            MeshData md = this.meshDataList.get(selectedItem);
            if (md.getCoordinate() != null) {
                SizeInfo si = MeshUtil.calcSizeInfo(md);
                this.displaySizeInfoPanel.setSizeInfo(si);
                //this.inputSizeInfoPanel.setSizeInfo(si);
            }
            this.meshData = md;
        }
    }
    
    
    //標準コンバート用のパネル
    class StandardConvertPanel extends JPanel implements ActionListener{
        private GraphView graphView; 
        private BufferedImage imageData;
        
        public StandardConvertPanel()
        {
            makeLayout();
        }
        
        private void makeLayout() {
            this.imageData = new BufferedImage(32, 32, BufferedImage.TYPE_INT_RGB);
            this.graphView = new GraphView();
            this.graphView.setImage(this.imageData);

            JButton convertButton = new JButton("Convert");
            JButton saveButton = new JButton("Save PNG");
            convertButton.setActionCommand("convert");
            saveButton.setActionCommand("save");
            convertButton.addActionListener(this);
            saveButton.addActionListener(this);
            
            GridBagLayout gbl = new GridBagLayout();
            GridBagConstraints gbc = new GridBagConstraints();
            gbc.gridx = 0;
            gbc.gridy = 0;
            gbc.gridwidth = 1;
            gbc.gridheight = 4;
            gbc.fill = GridBagConstraints.BOTH;
            gbc.weightx = 1.0;
            gbc.weighty = 1.0;
            gbl.setConstraints(graphView, gbc);
            gbc.gridx++;
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 0.0;
            gbc.weighty = 0.0;
            gbc.gridheight = 1;
            gbl.setConstraints(convertButton, gbc);
            gbc.gridy++;
            gbl.setConstraints(saveButton, gbc);

            setLayout(gbl);
            add(graphView);
            add(convertButton);
            add(saveButton);
        }
        public void actionPerformed(ActionEvent e)
        {
            try {
                if (e.getActionCommand().equals("convert")) {
                    if (ConvertToSculptPanel.this.meshData == null) {
                        JOptionPane.showMessageDialog(this, "MeshData is not loaded.");
                    } else if (dipRenderingConf.isValidData()) {
                        Object[] data1 = dipRenderingConf.getValues();
                        ConvertToSculptPanel.this.renderingConf.setVertexNumX((Integer)data1[0]);
                        ConvertToSculptPanel.this.renderingConf.setVertexNumY((Integer)data1[1]);
                        ConvertToSculptPanel.this.renderingConf.setImageWidth((Integer)data1[2]);
                        ConvertToSculptPanel.this.renderingConf.setImageHeight((Integer)data1[3]);

                        //sizeInfoが取得できない場合は自動正規化を行う
                        SizeInfo sizeInfo = ConvertToSculptPanel.this.inputSizeInfoPanel.getSizeInfo();
                        this.convert(ConvertToSculptPanel.this.meshData, ConvertToSculptPanel.this.renderingConf, sizeInfo);
                    } else {
                        JOptionPane.showMessageDialog(this, "Inputed data is invalid.");
                    }
                } else if (e.getActionCommand().equals("save")) {
                    if (this.imageData != null) {
                        if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
                            File outputFile = fileChooser.getSelectedFile();
                            {
                                String fname = outputFile.getAbsolutePath();
                                if (fname.length() > 4) {
                                    String extention = fname.substring(fname.length() - 4, fname.length()).toLowerCase(); 
                                    if (! extention.equals(".png")) {
                                        outputFile = new File(fname + ".png");
                                    }
                                }
                            }
                            if (outputFile.exists()) {
                                String msg = "A file named \"" + outputFile.getName() + "\" already exists.  Do you want to replace it?";
                                if (JOptionPane.showConfirmDialog(this, msg, "Select",JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
                                    ImageIO.write(imageData, "png", outputFile);
                                }
                            } else {
                                ImageIO.write(imageData, "png", outputFile);
                            }
                        }
                    }
                }
            } catch (Exception ex) {
                JOptionPane.showMessageDialog(this, ex.getMessage());
                ex.printStackTrace();
            }
        }
        /**
         * 
         * @param md                メッシュ情報 
         * @param rc                レンダリングの設定
         * @param sizeInfo          サイズの設定、nullの場合は自動正規化を行う
         */
        private void convert(MeshData md, RenderingConf rc, SizeInfo sizeInfo) {
            BufferedImage tmp = new BufferedImage(rc.getVertexNumX(), rc.getVertexNumY(), BufferedImage.TYPE_INT_RGB);
            BufferedImage out = new BufferedImage(rc.getImageWidth(), rc.getImageHeight(), BufferedImage.TYPE_INT_RGB);

            //メッシュデータの作成とチェック
            MeshData tmd = new MeshData(md);
            if (tmd.getCoordIndex() == null) {
                //CoordIndexが無いならばエラー
                JOptionPane.showMessageDialog(this, tmd.getName() + " does not contain coordIndex.");
                return;
            } else if (tmd.getCoordinate() == null) {
                //Coordinateが無いならばエラー
                JOptionPane.showMessageDialog(this, tmd.getName() + " does not contain coordinate.");
                return;
            } else if (tmd.getTexCoordIndex() == null) {
                //UV情報を持っていないならばエラー
                JOptionPane.showMessageDialog(this, tmd.getName() + " does not contain texCoordIndex.");
                return;
            } else if (tmd.getTexCoordinate() == null) {
                //UV情報を持っていないならばエラー
                JOptionPane.showMessageDialog(this, tmd.getName() + " does not contain texCoordinate.");
                return;
            }

            //サイズ情報が与えられているならば使用する。
            //無ければ自動正規化する
            if (sizeInfo != null) {
                MeshUtil.normalizeCoordinate(tmd, sizeInfo);
            } else {
                MeshUtil.autoNormalizeCoordinate(tmd);
            }
            
            MeshUtil.drawMesh(tmd, tmp, rc.getVertexNumX(), rc.getVertexNumY());
            
            int oX = (int)Math.round((rc.getImageWidth()  / (rc.getVertexNumX()-1))/2d);
            int oY = (int)Math.round((rc.getImageHeight() / (rc.getVertexNumY()-1))/2d);

            Graphics g = out.getGraphics();
            g.drawImage(tmp, -oX, out.getHeight()+oY, out.getWidth()+oX*2, -out.getHeight()-oY*2, null);
            g.dispose();
            
            this.imageData = out;
            this.graphView.setImage(out);
            
            validate();
            repaint();
        }
    }

    //強制コンバート用のパネル
    class ForceConvertPanel extends JPanel implements ActionListener{
        private GraphView graphView; 
        private BufferedImage imageData;
        private MeshData convertedMeshData;
        private InputVector3dPanel inputCenterPanel;
        private InputVector3dPanel inputAxisPanel;
        
        public ForceConvertPanel()
        {
            makeLayout();
        }
        
        private void makeLayout() {
            this.imageData = new BufferedImage(32, 32, BufferedImage.TYPE_INT_RGB);
            this.graphView = new GraphView();
            this.graphView.setImage(this.imageData);
            this.inputCenterPanel = new InputVector3dPanel("Center point", "Use SizeInfo's center");
            this.inputAxisPanel = new InputVector3dPanel("Rotate axis", "Use default");
            
            JButton convertButton = new JButton("Convert");
            JButton saveButton = new JButton("Save PNG");
            JButton saveVrmlButton = new JButton("Save VRML 1.0");
            JButton saveX3dButton = new JButton("Save X3D");
            JButton saveObjButton = new JButton("Save OBJ");
            convertButton.setActionCommand("convert");
            saveButton.setActionCommand("save");
            saveVrmlButton.setActionCommand("save vrml");
            saveX3dButton.setActionCommand("save x3d");
            saveObjButton.setActionCommand("save obj");
            convertButton.addActionListener(this);
            saveButton.addActionListener(this);
            saveVrmlButton.addActionListener(this);
            saveX3dButton.addActionListener(this);
            saveObjButton.addActionListener(this);
            
            GridBagLayout gbl = new GridBagLayout();
            GridBagConstraints gbc = new GridBagConstraints();
            gbc.gridx = 0;
            gbc.gridy = 0;
            gbc.gridwidth = 1;
            gbc.gridheight = 1;
            
            gbc.gridwidth = 1;
            gbc.gridheight = 6;
            gbc.fill = GridBagConstraints.BOTH;
            gbc.weightx = 1.0;
            gbc.weighty = 1.0;
            gbl.setConstraints(graphView, gbc);
            gbc.gridx++;
            
            gbc.weightx = 0.0;
            gbc.weighty = 0.0;
            gbc.gridheight = 5;
            gbl.setConstraints(inputCenterPanel, gbc);
            gbc.gridx++;
            gbl.setConstraints(inputAxisPanel, gbc);
            gbc.gridx++;

            
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 0.0;
            gbc.weighty = 0.0;
            gbc.gridheight = 1;
            gbl.setConstraints(convertButton, gbc);
            gbc.gridy++;
            gbl.setConstraints(saveButton, gbc);
            gbc.gridy++;
            gbl.setConstraints(saveVrmlButton, gbc);
            gbc.gridy++;
            gbl.setConstraints(saveX3dButton, gbc);
            gbc.gridy++;
            gbl.setConstraints(saveObjButton, gbc);

            setLayout(gbl);
            add(inputCenterPanel);
            add(inputAxisPanel);
            add(graphView);
            add(convertButton);
            add(saveButton);
            add(saveVrmlButton);
            add(saveX3dButton);
            add(saveObjButton);
        }
        public void actionPerformed(ActionEvent e)
        {
            try {
                if (e.getActionCommand().equals("convert")) {
                    if (ConvertToSculptPanel.this.meshData == null) {
                        JOptionPane.showMessageDialog(this, "MeshData is not loaded.");
                    } else if (dipRenderingConf.isValidData()) {
                        Object[] data1 = dipRenderingConf.getValues();
                        ConvertToSculptPanel.this.renderingConf.setVertexNumX((Integer)data1[0]);
                        ConvertToSculptPanel.this.renderingConf.setVertexNumY((Integer)data1[1]);
                        ConvertToSculptPanel.this.renderingConf.setImageWidth((Integer)data1[2]);
                        ConvertToSculptPanel.this.renderingConf.setImageHeight((Integer)data1[3]);

                        //sizeInfoが取得できない場合は自動正規化を行う
                        SizeInfo sizeInfo = ConvertToSculptPanel.this.inputSizeInfoPanel.getSizeInfo();
                        this.convert(ConvertToSculptPanel.this.meshData, ConvertToSculptPanel.this.renderingConf, sizeInfo);
                    } else {
                        JOptionPane.showMessageDialog(this, "Inputed data is invalid.");
                    }
                } else if (e.getActionCommand().equals("save")) {
                    if (this.imageData != null) {
                        File outputFile = getSaveTargetFile(".png");
                        if (outputFile != null) {
                            ImageIO.write(imageData, "png", outputFile);
                        }
                    }
                } else if (e.getActionCommand().equals("save vrml")) {
                    if (this.convertedMeshData != null) {
                        File outputFile = getSaveTargetFile(".wrl");
                        if (outputFile != null) {
                            FileOutputStream out = new FileOutputStream(outputFile);
                            meshConverterMap.get("vrml").writeMeshData(this.convertedMeshData, out);
                            out.close();
                        }
                    }
                } else if (e.getActionCommand().equals("save x3d")) {
                    if (this.convertedMeshData != null) {
                        File outputFile = getSaveTargetFile(".x3d");
                        if (outputFile != null) {
                            FileOutputStream out = new FileOutputStream(outputFile);
                            meshConverterMap.get("x3d").writeMeshData(this.convertedMeshData, out);
                            out.close();
                        }
                    }
                } else if (e.getActionCommand().equals("save obj")) {
                    if (this.convertedMeshData != null) {
                        File outputFile = getSaveTargetFile(".obj");
                        if (outputFile != null) {
                            FileOutputStream out = new FileOutputStream(outputFile);
                            meshConverterMap.get("obj").writeMeshData(this.convertedMeshData, out);
                            out.close();
                        }
                    }
                }
            } catch (Exception ex) {
                JOptionPane.showMessageDialog(this, ex.getMessage());
                ex.printStackTrace();
            }
        }
        /**
         * 
         * @param md                メッシュ情報 
         * @param rc                レンダリングの設定
         * @param sizeInfo          サイズの設定、nullの場合は自動正規化を行う
         */
        private void convert(MeshData md, RenderingConf rc, SizeInfo sizeInfo) {
            BufferedImage tmp = new BufferedImage(rc.getVertexNumX(), rc.getVertexNumY(), BufferedImage.TYPE_INT_RGB);
            BufferedImage out = new BufferedImage(rc.getImageWidth(), rc.getImageHeight(), BufferedImage.TYPE_INT_RGB);

            //メッシュデータの作成とチェック
            if (md.getCoordIndex() == null) {
                //CoordIndexが無いならばエラー
                JOptionPane.showMessageDialog(this, md.getName() + " does not contain coordIndex.");
                return;
            } else if (md.getCoordinate() == null) {
                //Coordinateが無いならばエラー
                JOptionPane.showMessageDialog(this, md.getName() + " does not contain coordinate.");
                return;
            }
            
            BVector3d center = this.inputCenterPanel.getVector3d();
            if (center == null)
            {
                center = new BVector3d();
                SizeInfo si;
                if (sizeInfo != null) {
                    si = sizeInfo;
                } else {
                    si = MeshUtil.calcSizeInfo(md);
                }
                center.x = (si.getMinX() + si.getMaxX()) / 2.0;
                center.y = (si.getMinY() + si.getMaxY()) / 2.0;
                center.z = (si.getMinZ() + si.getMaxZ()) / 2.0;
            }
            
            BVector3d axisAngle = inputAxisPanel.getVector3d();
            if (axisAngle == null) {
                axisAngle = new BVector3d();
            } else {
                axisAngle.scale(Math.PI / 180);
            }
            
            int numX;
            int numY;
            numX = rc.getVertexNumX()-1;
            numY = rc.getVertexNumY();
            BVector3d[][] vertices;
            {
                vertices = new BVector3d[numX][numY];
                for (int i=0;i<numX;i++) {
                    for(int j=0;j<numY;j++) {
                        vertices[i][j] = new BVector3d();
                    }
                }
            }
            
            //頂点配列からメッシュデータに変換。
            BMath.calcSphereMap(vertices, numX, numY, md, center, axisAngle);
            MeshData convertedMeshData = MeshUtil.convertBVector3dToMeshData(md.getName(), vertices, numX, numY, true);
            MeshData tmpMeshData = new MeshData(convertedMeshData);
            
            //サイズ情報が与えられているならば使用する。
            //無ければ自動正規化する
            if (sizeInfo != null) {
                MeshUtil.normalizeCoordinate(tmpMeshData, sizeInfo);
            } else {
                MeshUtil.autoNormalizeCoordinate(tmpMeshData);
            }
            MeshUtil.drawMesh(tmpMeshData, tmp, rc.getVertexNumX(), rc.getVertexNumY());
            
            int oX = (int)Math.round((rc.getImageWidth()  / (rc.getVertexNumX()-1))/2d);
            int oY = (int)Math.round((rc.getImageHeight() / (rc.getVertexNumY()-1))/2d);

            Graphics g = out.getGraphics();
            g.drawImage(tmp, -oX, out.getHeight()+oY, out.getWidth()+oX*2, -out.getHeight()-oY*2, null);
            g.dispose();
            
            this.imageData = out;
            this.graphView.setImage(out);
            this.convertedMeshData = convertedMeshData;
            
            validate();
            repaint();
        }
    }

    class FileSelectPanel extends JPanel {
        JTextField fieldFileName;
        public FileSelectPanel()
        {
            makeLayout();
        }

        private void makeLayout() {
            JLabel labelInputFile = new JLabel("Input file");
            this.fieldFileName = new JTextField();
            this.fieldFileName.setEditable(false);

            JButton buttonLoadX3d = new JButton("Load X3D");
            buttonLoadX3d.setActionCommand("load x3d");
            buttonLoadX3d.addActionListener(ConvertToSculptPanel.this);
            JButton buttonLoadVrml = new JButton("Load VRML");
            buttonLoadVrml.setActionCommand("load vrml");
            buttonLoadVrml.addActionListener(ConvertToSculptPanel.this);
            JButton buttonLoadObj = new JButton("Load Obj");
            buttonLoadObj.setActionCommand("load obj");
            buttonLoadObj.addActionListener(ConvertToSculptPanel.this);
            
            GridBagConstraints gbc = new GridBagConstraints();
            GridBagLayout gbl = new GridBagLayout();
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 0;
            gbc.weighty = 0;
            gbc.gridx = 0;
            gbc.gridy = 0;
            gbl.setConstraints(labelInputFile, gbc);
            gbc.gridx++;
            gbc.weightx = 1.0;
            gbl.setConstraints(this.fieldFileName, gbc);
            gbc.gridx++;
            gbc.weightx = 0;
            gbl.setConstraints(buttonLoadX3d, gbc);
            gbc.gridy++;
            gbl.setConstraints(buttonLoadVrml, gbc);
            gbc.gridy++;
            gbl.setConstraints(buttonLoadObj, gbc);
            
            setLayout(gbl);
            add(labelInputFile);
            add(this.fieldFileName);
            add(buttonLoadX3d);
            add(buttonLoadVrml);
            add(buttonLoadObj);
        }
        
        private void setFileName(String fileName) {
            this.fieldFileName.setText(fileName);
        }
    }
    
    public File getSaveTargetFile(String extension) {
        if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
            File outputFile = fileChooser.getSelectedFile();
            {
                String fname = outputFile.getAbsolutePath();
                if (fname.length() > extension.length()) {
                    String extention = fname.substring(fname.length() - extension.length(), fname.length()).toLowerCase(); 
                    if (! extention.equals(extension)) {
                        outputFile = new File(fname + extension);
                    }
                } else {
                    outputFile = new File(fname + extension);
                }
            }
            if (outputFile.exists()) {
                String msg = "A file named \"" + outputFile.getName() + "\" already exists.  Do you want to replace it?";
                if (JOptionPane.showConfirmDialog(this, msg, "Select",JOptionPane.YES_NO_OPTION) != JOptionPane.YES_OPTION) {
                    return null;
                }
            }
            return outputFile;
        }
        return null;
    }
}
